@@ -1,15 +1,25 @@
-#  Copyright (c) 2019 Data Spree UG (haftungsbeschraenkt) - All Rights Reserved
-#  Unauthorized copying of this file, via any medium is strictly prohibited.
-#  Proprietary and confidential.
+#  Copyright 2021 Data Spree GmbH
+#
+#  Licensed under the Apache License, Version 2.0 (the "License");
+#  you may not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+#       http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing, software
+#  distributed under the License is distributed on an "AS IS" BASIS,
+#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+#  See the License for the specific language governing permissions and
+#  limitations under the License.
 
 import json
 import logging
+import math
 import os
+import sys
 import threading
 
-import numpy as np
 import requests as rq
-from PIL import Image
 from joblib import Parallel, delayed
 from requests.adapters import HTTPAdapter
 from requests.auth import HTTPBasicAuth
@@ -32,12 +42,12 @@ class DLDSClient:
     :param auth_token: Platform authentication token.
     :param http_retries: Number of HTTP retries before error is thrown.
     :param parallel_requests: Maximum number of parallel HTTP request.
-    :param base_url: Base URL of the platform. Modify this parameter in case you use an on-premise installation.
+    :param server_url: URL of the DLDS server. Modify this parameter in case you use an on-premise installation.
 
     """
 
     def __init__(self, username=None, password=None, auth_token=None, http_retries=10, parallel_requests=16,
-                 base_url='https://api.vision.data-spree.com/api'):
+                 server_url='https://api.vision.data-spree.com/api', verify_ssl=True, verify_s3_ssl=True):
 
         self.username = username
         self.password = password
@@ -53,12 +63,16 @@ class DLDSClient:
             raise ValueError('Username and password or authentication token must be provided.')
 
         self.http_retries = http_retries
-        self.base_url = base_url
+        self.server_url = server_url
         self.parallel_requests = parallel_requests
         self.http = rq.Session()
+        if not verify_ssl:
+            self.http.verify = False
         self.s3_http = rq.Session()
+        if not verify_s3_ssl:
+            self.s3_http.verify = False
         self.http.auth = self.auth
-        retry = Retry(total=http_retries, connect=http_retries, backoff_factor=0.5,
+        retry = Retry(total=http_retries, connect=1, backoff_factor=0.5,
                       status_forcelist=[500, 502, 503, 504])
         adapter = HTTPAdapter(pool_maxsize=self.parallel_requests, max_retries=retry)
 
@@ -71,7 +85,20 @@ class DLDSClient:
                              self.auth_token,
                              self.http_retries,
                              self.parallel_requests,
-                             self.base_url,))
+                             self.server_url,
+                             self.http.verify,
+                             self.s3_http.verify,))
+
+    @staticmethod
+    def dlds_dir():
+        base_dir = os.path.expanduser('~')
+        dlds_dir = os.path.join(base_dir, '.dlds')
+        if sys.platform in ['win32', 'cygwin'] and os.getlogin() == 'SYSTEM':
+            # on windows in case the 'user' is SYSTEM, e.g. when started as a windows service
+            dlds_dir = os.path.join(os.getenv('ALLUSERSPROFILE'), 'Data Spree', 'Deep Learning DS')
+        if not os.path.exists(dlds_dir):
+            os.makedirs(dlds_dir)
+        return dlds_dir
 
     def download_dataset(self, directory, dataset_id, n_items=-1, accepted_status=None):
         """Download a dataset from data spree vision platform.
@@ -80,8 +107,8 @@ class DLDSClient:
         :param dataset_id: ID of the dataset that should be downloaded.
         :param n_items: Maximum number of items to download. Set to -1 to download all items.
         :param accepted_status: Download only items with the provided status (list of strings). Set to None to accept all status.
-
         """
+
         logger.info('Download dataset. Dataset ID: {}'.format(dataset_id))
 
         # download list of dataset items
@@ -150,7 +177,7 @@ class DLDSClient:
 
         # load item details
         try:
-            response = self.http.get(self.base_url + '/dataset-items/{}'.format(item_id))
+            response = self.http.get(self.server_url + '/dataset-items/{}'.format(item_id))
             response.raise_for_status()
         except rq.exceptions.RequestException as e:
             logger.error(e)
@@ -222,7 +249,7 @@ class DLDSClient:
         :return: Dictionary containing information of the requested subset.
         """
         try:
-            response = self.http.get(self.base_url + '/data-subsets/{}/'.format(subset_id))
+            response = self.http.get(self.server_url + '/data-subsets/{}/'.format(subset_id))
             if response.status_code == 200:
                 subset = response.json()
                 return subset
@@ -272,7 +299,7 @@ class DLDSClient:
         :return: Dictionary containing information of the requested datatset.
         """
         try:
-            response = self.http.get(self.base_url + '/datasets/{}/'.format(dataset_id))
+            response = self.http.get(self.server_url + '/datasets/{}/'.format(dataset_id))
             if response.status_code == 200:
                 dataset = response.json()
                 return dataset
@@ -339,7 +366,7 @@ class DLDSClient:
             dataset['semantic_segmentation_class_label_ids'] = semantic_segmentation_class_label_ids
 
         try:
-            response = self.http.post(self.base_url + '/datasets/', data=dataset)
+            response = self.http.post(self.server_url + '/datasets/', data=dataset)
             response.raise_for_status()
             data = response.json()
             return data['id']
@@ -361,13 +388,13 @@ class DLDSClient:
 
         n_items = subset['total_items']
         items_per_request = 5000
-        n_requests = int(np.ceil(n_items / items_per_request))
+        n_requests = int(math.ceil(n_items / items_per_request))
 
         all_items = []
         for i in range(n_requests):
             offset = i * items_per_request
-            url = self.base_url + '/data-subsets/{}/items/?limit={}&offset={}'.format(subset_id, items_per_request,
-                                                                                      offset)
+            url = self.server_url + '/data-subsets/{}/items/?limit={}&offset={}'.format(subset_id, items_per_request,
+                                                                                        offset)
             logger.debug('Get datasubset items from "{}"'.format(url))
             try:
                 response = self.http.get(url)
@@ -393,12 +420,13 @@ class DLDSClient:
 
         n_items = dataset['total_items']
         items_per_request = 5000
-        n_requests = int(np.ceil(n_items / items_per_request))
+        n_requests = int(math.ceil(n_items / items_per_request))
 
         all_items = []
         for i in range(n_requests):
             offset = i * items_per_request
-            url = self.base_url + '/datasets/{}/items/?limit={}&offset={}'.format(dataset_id, items_per_request, offset)
+            url = self.server_url + '/datasets/{}/items/?limit={}&offset={}'.format(dataset_id, items_per_request,
+                                                                                    offset)
             logger.debug('Get dataset items from "{}"'.format(url))
             try:
                 response = self.http.get(url)
@@ -444,7 +472,7 @@ class DLDSClient:
         }
 
         try:
-            response = self.http.post(self.base_url + '/dataset-items/', data=dataset_item, files=files)
+            response = self.http.post(self.server_url + '/dataset-items/', data=dataset_item, files=files)
             response.raise_for_status()
             data = response.json()
             return data['id']
@@ -473,7 +501,7 @@ class DLDSClient:
         """
         # load item details
         try:
-            response = self.http.get(self.base_url + '/dataset-items/{}'.format(item_id))
+            response = self.http.get(self.server_url + '/dataset-items/{}'.format(item_id))
             response.raise_for_status()
         except rq.exceptions.RequestException as e:
             logger.error(e)
@@ -509,7 +537,7 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.get(self.base_url + '/models/{}/'.format(model_id))
+            response = self.http.get(self.server_url + '/models/{}/'.format(model_id))
             if response.status_code == 200:
                 model = response.json()
                 return model
@@ -534,7 +562,7 @@ class DLDSClient:
         :return: List of class labels (ID, name).
         """
         try:
-            response = self.http.get(self.base_url + '/models/{}/class_labels/'.format(model_id))
+            response = self.http.get(self.server_url + '/models/{}/class_labels/'.format(model_id))
             if response.status_code == 200:
                 model = response.json()
                 return model
@@ -574,7 +602,7 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.get(self.base_url + '/models/{}/parameters/'.format(model_id))
+            response = self.http.get(self.server_url + '/models/{}/parameters/'.format(model_id))
             if response.status_code == 200:
                 model = response.json()
                 return model
@@ -598,7 +626,7 @@ class DLDSClient:
         :return: List with checkpoints.
         """
         try:
-            response = self.http.get(self.base_url + '/models/{}/checkpoints/'.format(model_id))
+            response = self.http.get(self.server_url + '/models/{}/checkpoints/'.format(model_id))
             if response.status_code == 200:
                 checkpoints = response.json()
                 return checkpoints
@@ -626,7 +654,7 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.get(self.base_url + '/cloud-deployments/{}/'.format(id))
+            response = self.http.get(self.server_url + '/cloud-deployments/{}/'.format(id))
             if response.status_code == 200:
                 return response.json()
         except rq.exceptions.RequestException as e:
@@ -672,7 +700,7 @@ class DLDSClient:
         logger.info('Download model checkpoint with ID: {}'.format(checkpoint_id))
 
         try:
-            response = self.http.get(self.base_url + '/model-checkpoints/' + str(checkpoint_id))
+            response = self.http.get(self.server_url + '/model-checkpoints/' + str(checkpoint_id))
             response.raise_for_status()
             checkpoint = response.json()
         except rq.exceptions.RequestException as e:
@@ -710,7 +738,7 @@ class DLDSClient:
         }
 
         try:
-            response = self.http.post(self.base_url + '/model-checkpoints/', data=model_checkpoint)
+            response = self.http.post(self.server_url + '/model-checkpoints/', data=model_checkpoint)
             response.raise_for_status()
             return response.json()
         except rq.exceptions.RequestException as e:
@@ -736,7 +764,7 @@ class DLDSClient:
         }
 
         try:
-            response = self.http.patch(self.base_url + '/model-checkpoints/{}/'.format(checkpoint_id),
+            response = self.http.patch(self.server_url + '/model-checkpoints/{}/'.format(checkpoint_id),
                                        data=model_checkpoint, files=files)
             response.raise_for_status()
             logger.info('Model checkpoint uploaded: {}'.format(file_name))
@@ -795,7 +823,7 @@ class DLDSClient:
         files = {'inference_data': (file_name, open(file_path, 'rb'))}
 
         try:
-            response = self.http.patch(self.base_url + '/models/{}/'.format(model_id), files=files)
+            response = self.http.patch(self.server_url + '/models/{}/'.format(model_id), files=files)
             response.raise_for_status()
         except rq.exceptions.RequestException as e:
             logger.error(e)
@@ -830,7 +858,7 @@ class DLDSClient:
         }
 
         try:
-            response = self.http.post(self.base_url + '/model-statistics/', data=statistics_entry)
+            response = self.http.post(self.server_url + '/model-statistics/', data=statistics_entry)
             response.raise_for_status()
         except rq.exceptions.RequestException as e:
             logger.error(e)
@@ -865,7 +893,8 @@ class DLDSClient:
         >>>             'class_confidences': {...}  # dictionary representing the confidences (class_id -> confidence)
         >>>         },
         >>>         'detection': {
-        >>>             'boxes': [...]              # list of arrays representing the box coordinates and position [x, y, width, height, orientation (optional)]
+        >>>             'boxes': [...]              # list of arrays representing the box coordinates and positions [x, y, width, height, orientation (optional)]
+        >>>                                         # x, y, width, and height must be scaled to the interval [0, 1] (the point (1, 1) corresponds to the lower right corner)
         >>>             'class_confidences': {...}  # dictionary representing the detection confidences (class_id -> confidence)
         >>>         },
         >>>     }
@@ -892,7 +921,7 @@ class DLDSClient:
         if override_latest:
             # load list of existing inference results
             try:
-                response = self.http.get(self.base_url + '/models/{}/latest_inference_results/'.format(model_id))
+                response = self.http.get(self.server_url + '/models/{}/latest_inference_results/'.format(model_id))
                 if response.status_code == 200:
                     existing = response.json()
                     for result in existing:
@@ -926,12 +955,12 @@ class DLDSClient:
                     box = result['detection']['boxes'][j]
 
                     class_confidences = {}
+
                     class_confidences[0] = float(result['detection']['class_confidences'][j][0])  # background class
                     for c in range(1, len(result['detection']['class_confidences'][j])):
                         class_id = inverse_class_id_map[c]
                         class_confidences[class_id] = float(result['detection']['class_confidences'][j][c])
 
-                    # confidence = inverse_class_id_map[result['class_ids'][j]]
                     detection_result.append({
                         'id': j,
                         'class_confidences': class_confidences,
@@ -959,7 +988,7 @@ class DLDSClient:
                     'result': json.dumps(result_dict)
                 }
 
-                requests.append((self.http.post, '{}/inference-results/'.format(self.base_url), inference_result))
+                requests.append((self.http.post, '{}/inference-results/'.format(self.server_url), inference_result))
 
             else:
                 inference_result = {
@@ -969,7 +998,7 @@ class DLDSClient:
                 }
 
                 requests.append((self.http.patch,
-                                 '{}/inference-results/{}/'.format(self.base_url, existing_result['id']),
+                                 '{}/inference-results/{}/'.format(self.server_url, existing_result['id']),
                                  inference_result))
 
         if self.parallel_requests > 1:
@@ -995,7 +1024,7 @@ class DLDSClient:
         :return: List of class labels or `None` in case of errors.
         """
         try:
-            response = self.http.get(self.base_url + '/class-labels/')
+            response = self.http.get(self.server_url + '/class-labels/')
             response.raise_for_status()
             data = response.json()
             return data
@@ -1014,7 +1043,7 @@ class DLDSClient:
         class_label = {'name': name}
 
         try:
-            response = self.http.post(self.base_url + '/class-labels/', data=class_label)
+            response = self.http.post(self.server_url + '/class-labels/', data=class_label)
             response.raise_for_status()
             data = response.json()
             return data
@@ -1031,7 +1060,7 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.get(self.base_url + '/workers/{}/jobs/'.format(worker_id))
+            response = self.http.get(self.server_url + '/workers/{}/jobs/'.format(worker_id))
             response.raise_for_status()
             data = response.json()
             return data
@@ -1048,7 +1077,7 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.get(self.base_url + '/jobs/{}/'.format(job_id))
+            response = self.http.get(self.server_url + '/jobs/{}/'.format(job_id))
             response.raise_for_status()
             data = response.json()
             return data
@@ -1069,7 +1098,7 @@ class DLDSClient:
             return None
 
         try:
-            response = self.http.patch(self.base_url + '/jobs/{}/'.format(job_id), data=job)
+            response = self.http.patch(self.server_url + '/jobs/{}/'.format(job_id), data=job)
             response.raise_for_status()
             data = response.json()
             return data
@@ -1085,8 +1114,11 @@ class DLDSClient:
         """
 
         try:
-            response = self.http.delete(self.base_url + '/jobs/{}/'.format(job_id))
-            response.raise_for_status()
+            response = self.http.delete(self.server_url + '/jobs/{}/'.format(job_id))
+            if response.status_code == 404:
+                logger.info('Job could not be deleted because it did not exist.')
+            else:
+                response.raise_for_status()
         except rq.exceptions.RequestException as e:
             logger.error(e)
 
@@ -1214,6 +1246,27 @@ class DLDSClient:
             class_label_map[class_label['name']] = class_label['id']
         return class_label_map
 
+    def create_worker(self, name):
+        """Create a new training worker.
+
+        :return: ID of the newly created worker or `None`.
+        """
+
+        try:
+            worker = {
+                'name': name,
+                'status': '',
+                'details': {}
+            }
+            response = self.http.post(self.server_url + '/workers/', data=worker)
+            response.raise_for_status()
+            data = response.json()
+            return data
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return None
+
     def import_dlds(self, dataset_name, dataset_id, images, annotations):
         """Import a dataset that is available in the DLDS format.
 
@@ -1258,7 +1311,10 @@ class DLDSClient:
 
         # upload dataset items
         annotation_files = os.listdir(annotations)
+        annotation_files.sort()
+
         image_files = os.listdir(images)
+        image_files.sort()
 
         logger.info('Upload dataset items')
         for image_file in tqdm(image_files):
@@ -1274,7 +1330,7 @@ class DLDSClient:
             base_name_split = base_name.split('_')
 
             # first two elements of base_name_split contain dataset_id and item_id
-            image_file_name = '{}.{}'.format('_'.join(base_name_split[2:]), image_extension)
+            image_file_name = '{}{}'.format('_'.join(base_name_split[2:]), image_extension)
 
             annotation = {}
             # load annotations if they exist
@@ -1348,8 +1404,6 @@ class DLDSClient:
                 if image_file in dataset_item_names:
                     continue
 
-                base_name = os.path.splitext(image_file)[0]
-
                 image_file_path = os.path.join(directory, subfolder, image_file)
 
                 if self.create_dataset_item(image_file_path, dataset_id, class_annotation) == -1:
@@ -1372,6 +1426,9 @@ class DLDSClient:
         """
         logger.info('Import KITTI data')
 
+        # we import Pillow here as it is not required for other endpoints
+        from PIL import Image
+
         # create new dataset if no id is given
         if dataset_id == -1:
             # iterate over all items and determine used class labels
